package lib.turtle;

import kernel.turtle.Types.ToolSide;
import kernel.log.Log;
import kernel.turtle.TurtleSlot;
import kernel.turtle.Types.InteractDirections;
import kernel.turtle.Turtle;

using Lambda;
using tink.CoreApi;

/**
	A wrapper for the native turtle inventory functions. Adds usefullt helper functions.
**/
class InvManager {
	public static function place(item:Item, dir:InteractDirections):Outcome<Noise, String> {
		var invState = new InvState();
		var slot = invState.getSlotWithMinCountForItem(item);
		if (slot == null) {
			return Failure("Item not in inventory");
		}

		Turtle.selectSlot(slot);
		return Turtle.place(dir);
	}

	/**
		Cleans up turtle inventory. Moves items together.
		This should be run in a turtle thread.
	**/
	public static function defrag() {
		var invState = new InvState();
		var items = invState.getSlotsForItems();

		for (item in items) {
			defragItem(invState, item);
		}
	}

	private static function defragItem(invState:InvState, slots:Array<TurtleSlot>) {
		// Sort slots by items inside.
		slots.sort((a, b) -> {
			return invState.get(a).count - invState.get(b).count;
		});

		var maxInSlot = invState.get(slots[0]).max;

		// Loop from both sides. Move the slots with the lowest count into the fullest.
		var topIndex = slots.length - 1;

		for (k => slot in slots) {
			if (topIndex < k) {
				return;
			}

			var count = invState.get(slot).count;

			if (count == maxInSlot) {
				return;
			}

			var loopProtect = 20;

			while (loopProtect > 0) {
				loopProtect--;

				if (topIndex < k) {
					return;
				}

				var topCount = invState.get(slots[topIndex]).count;
				var spaceInTop = maxInSlot - topCount;

				if (spaceInTop == 0) {
					topIndex--;
					continue;
				}

				var transferCount = MathI.min(count, spaceInTop);
				if (!Turtle.transfer(slot, slots[topIndex], transferCount).isSuccess()) {
					Log.error("Failed to transfer items");
				}

				var topState = invState.get(slots[topIndex]);
				topState.count += transferCount;
				invState.set(slots[topIndex], topState);

				var botState = invState.get(slot);
				botState.count -= transferCount;
				invState.set(slot, botState);

				count -= transferCount;

				if (count <= 0) {
					break;
				}
			}

			if (loopProtect < 0) {
				Log.silly("Loopprotection triggerd");
			}
		}
	}

	public static function equipTool(item:Item, side:ToolSide):Outcome<TurtleSlot, String> {
		var invState = new InvState();
		var toolSlot = invState.findItemSlot(item);

		if (toolSlot == null) {
			return Failure("Item not found");
		}

		Turtle.selectSlot(toolSlot);
		var result = Turtle.equip(side);

		switch result {
			case Success(_):
				return Success(toolSlot);
			case Failure(failure):
				return Failure(failure);
		}
	}
}
